#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Standard Library
import logging

from contextlib import AbstractContextManager
from contextlib import nullcontext as does_not_raise
from typing import Generator, Optional

# 3rd Party Library
import pytest
import pytest_twisted

from scrapy.crawler import Crawler
from scrapy.http.request import Request
from scrapy.http.response.html import HtmlResponse
from scrapy.settings import Settings, iter_default_settings

# 1st Party Library
from scrachy.http_ import SeleniumRequest
from scrachy.middleware.selenium import AsyncSeleniumMiddleware
from scrachy.settings.defaults import selenium as selenium_defaults
from scrachy.settings.defaults.selenium import WebDriverName
from tests.integration.middleware.conftest import MockSpider, SettingsFactory
from tests.integration.middleware.test_selenium import paginate

log = logging.getLogger("test_async_selenium")

pytest.skip(
    "The async middleware is probably broken and difficult to fix",
    allow_module_level=True,
)


@pytest.fixture(scope="module")
def middleware() -> Generator[AsyncSeleniumMiddleware, None, None]:
    settings = Settings(dict(iter_default_settings()))
    settings.setmodule(selenium_defaults)
    settings.set("CONCURRENT_REQUESTS", 4)  # temporary
    settings.set("SCRACHY_SELENIUM_WEB_DRIVER", "Chrome")
    settings.set("SCRACHY_SELENIUM_WEB_DRIVER_OPTIONS", ["--headless=new"])

    crawler = Crawler(spidercls=MockSpider, settings=settings)
    middleware = AsyncSeleniumMiddleware.from_crawler(crawler)

    yield middleware

    log.info("Calling spider_closed")
    middleware.spider_closed(crawler.spider)


@pytest_twisted.ensureDeferred
@pytest.mark.parametrize(
    "driver, options, extensions, raises_expectation",
    [
        (None, None, None, does_not_raise()),
        ("Chrome", ["--headless=new"], True, does_not_raise()),
        ("Firefox", ["-headless"], True, does_not_raise()),
    ],
    indirect=["extensions"],
)
async def test_initialize_driver(
    settings_factory: SettingsFactory,
    driver: WebDriverName,
    options: Optional[list[str]],
    extensions: Optional[list[str]],
    raises_expectation: AbstractContextManager,
):
    with raises_expectation:
        settings = settings_factory(driver, options, extensions)

        crawler = Crawler(spidercls=MockSpider, settings=settings)

        try:
            middleware = AsyncSeleniumMiddleware.from_crawler(crawler)
            assert middleware.drivers.qsize() == settings.getint("CONCURRENT_REQUESTS")
        except:  # noqa: E722
            middleware = None
        finally:
            if middleware is not None:
                middleware.spider_closed(crawler.spider)


@pytest_twisted.ensureDeferred
async def test_return_none_on_scrapy_request(middleware: AsyncSeleniumMiddleware):
    scrapy_request = Request(url="http://not-an-url")

    assert middleware.process_request(scrapy_request) is None


@pytest_twisted.ensureDeferred
async def test_return_response(middleware: AsyncSeleniumMiddleware):
    request = SeleniumRequest(url="https://www.scrapethissite.com/pages/forms/")
    result = middleware.process_request(request)

    assert result is not None

    response = await result

    assert isinstance(response, HtmlResponse)

    title = response.css("title::text").get()
    assert title is not None
    assert "Hockey Teams" in title


@pytest_twisted.ensureDeferred
async def test_screenshot(middleware: AsyncSeleniumMiddleware):
    request = SeleniumRequest(
        url="https://www.scrapethissite.com/pages/forms/", screenshot=True
    )
    result = middleware.process_request(request)

    assert result is not None

    response = await result

    assert isinstance(response, HtmlResponse)
    assert response.meta["screenshot"] is not None


@pytest_twisted.ensureDeferred
async def test_script(middleware: AsyncSeleniumMiddleware):
    request = SeleniumRequest(
        url="https://www.scrapethissite.com/pages/ajax-javascript/",
        script_executor=paginate,
    )

    result = middleware.process_request(request)
    assert result is not None
    response = await result
    assert isinstance(response, HtmlResponse)

    expected_titles = {
        "2015": "Spotlight",
        "2014": "Birdman",
        "2013": "12 Years a Slave",
        "2012": "Argo",
        "2011": "The Artist",
        "2010": "The King's Speech",
    }

    for key, script_response in response.meta["script_result"].items():
        title = (
            script_response.css("tbody#table-body > tr.film > td.film-title::text")
            .get()
            .strip()
        )

        assert expected_titles[key] == title
        assert expected_titles[key] == title



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""Custom SqlAlchemy types."""

# Standard Library
from typing import Any

# 3rd Party Library
from sqlalchemy import JSON, Dialect, TypeDecorator
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.sql.type_api import TypeEngine


class ConditionalJson(TypeDecorator):
    """Uses JSONB if the dialect is postgresql otherwise uses JSON"""

    # The common base type
    impl = JSON

    def load_dialect_impl(self, dialect: Dialect) -> TypeEngine[Any]:
        if dialect.name == "postgresql":
            return dialect.type_descriptor(JSONB())

        if dialect.name in ("sqlite", "mysql", "mssql"):
            return dialect.type_descriptor(JSON())

        raise ValueError("The dialect must at least support the JSON data type.")



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Standard Library
# Python Modules
import datetime
import logging
import os
import pathlib

from typing import Any, Optional, TypeVar

# 3rd Party Library
# 3rd Party Modules
from dateutil import parser as date_parser
from dotenv import load_dotenv
from scrapy.settings import Settings
from sqlalchemy import create_engine
from sqlalchemy.exc import OperationalError

# 1st Party Library
# Project Modules
from scrachy.utils.db import construct_url

load_dotenv()

log = logging.getLogger(__name__)


T = TypeVar("T")

TEST_DATABASE_NAME = "scrachy_test"


def load_postgresql_settings() -> Settings:
    sslrootcert: Optional[str] = os.getenv("TEST_SCRACHY_DB_SSL_ROOT_CERT")
    sslcert: Optional[str] = os.getenv("TEST_SCRACHY_DB_SSL_CERT")
    sslkey: Optional[str] = os.getenv("TEST_SCRACHY_DB_SSL_KEY")
    sslmode: Optional[str] = os.getenv("TEST_SCRACHY_DB_SSL_MODE")

    connect_args = {
        "sslrootcert": sslrootcert,
        "sslcert": sslcert,
        "sslkey": sslkey,
        "sslmode": sslmode,
    }
    connect_args = {k: v for k, v in connect_args.items() if v}

    settings = Settings({
        "SCRACHY_DB_DIALECT": "postgresql",
        "SCRACHY_DB_DRIVER": "psycopg",
        "SCRACHY_DB_HOST": os.getenv("TEST_SCRACHY_DB_HOST", "localhost"),
        "SCRACHY_DB_PORT": int(os.getenv("TEST_SCRACHY_DB_PORT", "5432")),
        "SCRACHY_DB_DATABASE": os.getenv("TEST_SCRACHY_DB_DATABASE", "scrachy_test"),
        "SCRACHY_DB_USERNAME": os.getenv("TEST_SCRACHY_DB_USERNAME", "scrachy"),
        "SCRACHY_DB_PASSWORD": os.getenv("TEST_SCRACHY_DB_PASSWORD"),
        "SCRACHY_DB_CONNECT_ARGS": connect_args,
    })

    return settings


def is_postgresql_setup() -> bool:
    """
    Check to see if it is possible to connect to the ``scrachy_test``
    database using the user ``scrachy``.

    :return:
    """
    try:
        settings = load_postgresql_settings()
    except Exception as e:
        log.error(
            f"There was a problem configuring the test database settings: {str(e)}"
        )
        return False

    url = construct_url(settings)
    engine = create_engine(url, connect_args=settings["SCRACHY_DB_CONNECT_ARGS"])

    try:
        with engine.connect():
            can_connect = True
    except OperationalError as e:
        log.warning(f"Unable to connect to the postgresql test database: {str(e)}")
        can_connect = False
    finally:
        if engine is not None:
            engine.dispose()

    return can_connect


def update_database_settings(
    original_settings: Settings | list[Settings], dialect: str, basedir: pathlib.Path
):
    if isinstance(original_settings, Settings):
        original_settings = [original_settings]

    for current_settings in original_settings:
        current_settings["SCRACHY_DB_DIALECT"] = dialect

        if dialect in ("sqlite", "other"):
            db_dir = basedir
            db_dir.mkdir(parents=True, exist_ok=True)
            db_path = db_dir / f"{TEST_DATABASE_NAME}.sqlite3"
            db_path.unlink(missing_ok=True)

            current_settings.set("SCRACHY_DB_DIALECT", "sqlite")
            current_settings.set("SCRACHY_DB_HOST", None)
            current_settings.set("SCRACHY_DB_PORT", None)
            current_settings.set("SCRACHY_DB_USERNAME", None)
            current_settings.set("SCRACHY_DB_PASSWORD", None)
            current_settings.set("SCRACHY_DB_DATABASE", str(db_path))
            current_settings.set("SCRACHY_DB_SCHEMA", None)
            current_settings.set("SCRACHY_DB_CONNECT_ARGS", {})
        elif dialect == "postgresql":
            current_settings.update(load_postgresql_settings())


def parse_date(obj: Any) -> datetime.datetime:
    if isinstance(obj, str):
        return date_parser.parse(obj)

    if isinstance(obj, datetime.datetime):
        return obj

    raise ValueError(f"Can't parse: {str(obj)}")



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Standard Library
import copy
import logging

from collections import defaultdict

# 3rd Party Library
import pytest

from dotenv import load_dotenv
from integration.conftest import sync_engine_manager
from psycopg import Connection
from rwskit.sqlalchemy.engine import SyncAlchemyEngine
from rwskit.sqlalchemy.repository import ConflictResolutionStrategy
from scrapy.settings import Settings

# 1st Party Library
from scrachy.db.base import ScrachyBase
from scrachy.db.models import Response as CachedResponse
from scrachy.db.models import ScrapeHistory
from scrachy.db.repositories import SyncResponseRepository, SyncScrapeHistoryRepository

load_dotenv()


log = logging.getLogger("test_repositories")

log.debug("From test_repositories")


@pytest.fixture
def settings(
    settings_choices: dict[str, Settings], request: pytest.FixtureRequest
) -> Settings:
    return settings_choices[request.param]


# Note: It's difficult (maybe impossible?) to test if 'insert'/'upsert' is
# working without also testing 'find' and vice versa.
@pytest.mark.parametrize("dialect", ["sqlite", "postgresql"])
@pytest.mark.parametrize("settings", ["defaults", "settings_2"], indirect=True)
def test_upsert_and_find(
    dialect: str,
    settings: Settings,
    cached_responses: list[CachedResponse],
    postgresql: Connection,
):
    # Create a copy because the objects could be modified by the upsert
    # process.
    cached_responses_copy = copy.deepcopy(cached_responses)

    with sync_engine_manager(dialect, postgresql) as sa_engine:
        engine = SyncAlchemyEngine(sa_engine, ScrachyBase)

        with engine.session_scope():
            response_repo = SyncResponseRepository(engine)
            history_repo = SyncScrapeHistoryRepository(engine)

            for exp_response in cached_responses_copy:
                history = ScrapeHistory(
                    fingerprint=exp_response.fingerprint,
                    scrape_timestamp=exp_response.scrape_timestamp,
                    body=exp_response.body,
                )
                response_repo.upsert(
                    exp_response, on_conflict=ConflictResolutionStrategy.UPDATE
                )
                history_repo.insert(history)

                # Don't assert anything here. It's complicated with the scrape_history
                # assert act_response == exp_response

            with engine.session_scope():
                response_repo = SyncResponseRepository(engine)

                sorted_responses = sorted(
                    cached_responses,
                    key=lambda e: (e.fingerprint, e.scrape_timestamp),
                    reverse=True,
                )
                expected_responses = []
                fingers = defaultdict(list)

                for r in sorted_responses:
                    if r.fingerprint not in fingers:
                        r.scrape_timestamp = r.scrape_timestamp
                        expected_responses.append(r)

                    fingers[r.fingerprint].append(r.body)

                # One at a time
                # Compare strings to avoid comparing the scrape history

                assert all([
                    str(e) == str(response_repo.find_by_fingerprint(e.fingerprint))
                    for e in expected_responses
                ])

                # All of them
                with engine.session_scope() as session:
                    act_responses = response_repo.find_all_models(session=session)
                    assert len(expected_responses) == len(act_responses)
                    assert all([
                        str(e) == str(a)
                        for e, a in zip(expected_responses, act_responses)
                    ])

                    # Make sure the scrape history has the correct number of entries
                    for act_response in act_responses:
                        act_history = sorted(
                            act_response.scrape_history,
                            key=lambda h: h.scrape_timestamp,
                            reverse=True,
                        )
                        exp_history = fingers[act_response.fingerprint]

                        assert len(exp_history) == len(act_history)
                        assert all([
                            exp == act.body
                            for exp, act in zip(exp_history, act_history)
                        ])



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
"""
A simple server that receives pickled ``SeleniumRequests`` from ``stdin`` and
sends back pickled ``HtmlResponses`` to ``stdout``. It is meant to be used with
:class:`~scrachy.middleware.selenium.AsyncSeleniumMiddleware`.
"""

# Future Library
from __future__ import annotations

# Standard Library
import argparse
import logging
import pickle
import sys

from struct import pack, unpack
from typing import Optional, cast

# 3rd Party Library
from scrapy.http.request import Request
from scrapy.http.response.html import HtmlResponse
from selenium.common import TimeoutException
from selenium.webdriver.remote.webdriver import WebDriver

# 1st Party Library
from scrachy.http_ import SeleniumRequest
from scrachy.utils.selenium import (
    BufferIncompleteError,
    ShutdownRequest,
    UnhandledError,
    UnknownMessageType,
    initialize_driver,
    process_request,
)

log = logging.getLogger("driver_process")


Message = (
    HtmlResponse
    | pickle.PickleError
    | Request
    | SeleniumRequest
    | ShutdownRequest
    | TimeoutException
    | UnhandledError
    | UnknownMessageType
)


DEFAULT_BUFFER_SIZE = 256
REQUEST_HEADER_SIZE = 8


def strippable(arg: str) -> str:
    return arg.replace('"', "").strip()


def setup_logging(args: argparse.Namespace):
    if not args.log_file:
        return

    log_file = args.log_file
    logging.basicConfig(filename=log_file, level=logging.DEBUG, filemode="w")
    log.info("Logging is setup")


def decode_message(buffer: bytearray) -> Optional[Message]:
    # A message will always have two 4 byte header fields
    while len(buffer) >= REQUEST_HEADER_SIZE:
        req_length, msg_length = decode_header(buffer)

        if len(buffer) >= msg_length:
            # Get the data (minus the msg_length) and strip the padding
            data_start, data_end = REQUEST_HEADER_SIZE, REQUEST_HEADER_SIZE + req_length
            data = buffer[data_start:data_end]

            # Deserialize the object
            obj = pickle.loads(data)

            # Trim off the processed data
            buffer[:] = buffer[msg_length:]

            return obj

        # The data transmission is not complete
        raise BufferIncompleteError()


def decode_header(buffer: bytearray) -> tuple[int, int]:
    # The total length of data sent from one call. This should always
    # be the buffer_size.
    req_length = unpack("!I", buffer[:4])[0]

    # The length of the actual data.
    msg_length = unpack("!I", buffer[4:8])[0]

    return req_length, msg_length


def send_message(message: Message):
    data = pickle.dumps(message)
    res_length = pack("!I", len(data))
    sys.stdout.buffer.write(res_length + data)
    sys.stdout.buffer.flush()


def message_loop(driver: WebDriver, buffer: bytearray, buffer_size: int):
    while True:
        buffer.extend(sys.stdin.buffer.read(buffer_size))

        try:
            in_message: Optional[Message] = decode_message(buffer)

            if isinstance(in_message, ShutdownRequest):
                return

            if isinstance(in_message, SeleniumRequest):
                request = cast(SeleniumRequest, in_message)
                msg = process_request(driver, request)

                if msg is None:
                    raise ValueError(
                        f"Error processing input message '{in_message}'. Received a "
                        "null response when processing the request."
                    )

                send_message(msg)

            else:
                send_message(UnknownMessageType(str(type(in_message))))
        except pickle.PickleError as e:
            send_message(e)
        except TimeoutException as e:
            send_message(e)
        except BufferIncompleteError:
            pass  # Keep going
        except Exception as e:
            send_message(UnhandledError(e))


def main(args: argparse.Namespace):
    """
    The main driver program that starts the message loop.

    :param args:
    :return:
    """
    setup_logging(args)

    driver = None

    try:
        buffer = bytearray()
        driver = initialize_driver(
            args.driver, args.driver_options, args.driver_extensions
        )
        message_loop(driver, buffer, args.buffer_size)
    except Exception as e:
        send_message(UnhandledError(e))
    finally:
        if driver is not None:
            driver.quit()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()

    # It's not entirely clear why, but when the arguments are passed in via the
    # twisted spawnProcess quotes and extra spaces are added to the arguments.
    # This is especially problematic for the options, extensions, and log_file.
    # The `strippable` function removes all quotes from the parsed arguments
    # and strips any surrounding spaces. It is possible, although unlikely,
    # this could cause unintended consequences for some argument names or
    # paths.
    parser.add_argument(
        "-d", "--driver", choices=["Chrome", "Firefox"], default="Chrome"
    )
    parser.add_argument(
        "-o", "--driver-options", type=strippable, action="append", default=[]
    )
    parser.add_argument(
        "-e", "--driver-extensions", type=strippable, action="append", default=[]
    )
    parser.add_argument("-w", "--implicit_wait", type=int, default=None)
    parser.add_argument("-b", "--buffer-size", type=int, default=DEFAULT_BUFFER_SIZE)
    parser.add_argument("-f", "--log-file", type=strippable)
    parser.add_argument("-p", "--verify-proxy", action=argparse.BooleanOptionalAction)

    main(parser.parse_args())



##############################################################################
#  Copyright 2020 Reid Swanson and the Scrapy developers.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
##############################################################################

"""Hashing utilities."""

# Standard Library
from typing import Optional, Protocol, Sequence

# 3rd Party Library
from scrapy.utils.python import to_bytes


class Hasher(Protocol):
    """
    An object that can hash byte strings.
    """

    def update(self, b: bytes): ...

    def digest(self) -> bytes: ...


def hash_text(
    hash_fn: Hasher, text: Optional[str | Sequence[str | None]] = None
) -> bytes:
    """
    Hash a string, or list of strings, using the given ``hash_fn``. If any
    string is ``None``, it will return the hash value of an empty string.

    :param text: The string to hash.
    :param hash_fn: A hash function object that adheres to the hashlib API.
           Namely, it must have an ``update`` function and a ``digest``
           function.
    :return: The hash digest as an array of bytes.

    """
    if text is None or isinstance(text, str):
        text = [text]
    elif not isinstance(text, Sequence):
        raise ValueError(
            "If the input is not None, it must either be a string or sequence of "
            f"strings, but got: '{type(text)}'"
        )

    for t in text:
        hash_fn.update(to_bytes(t or b""))

    return hash_fn.digest()


# endregion Utility Methods



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Standard Library
import datetime
import logging
import re

from collections.abc import Iterable
from contextlib import contextmanager
from contextlib import nullcontext as does_not_raise

# 3rd Party Library
import arrow
import pytest

from integration.middleware.conftest import MockSpider
from rwskit.sqlalchemy.engine import SyncAlchemyEngine
from scrapy.crawler import Crawler
from scrapy.exceptions import IgnoreRequest
from scrapy.http.request import Request
from scrapy.settings import Settings, iter_default_settings
from sqlalchemy import delete

# 1st Party Library
from scrachy.db.models import Response
from scrachy.db.repositories import SyncResponseRepository
from scrachy.middleware.filter import CachedResponseFilter
from scrachy.settings.defaults.storage import PatternLike

log = logging.getLogger("test_cache_filter")


CACHED_URLS = [
    "https://www.google.com/page1",
    "https://www.google.com/page2",
    "https://university.edu",
    "https://university.edu/page1",
    "https://example.org",
]

UN_CACHED_URLS = [
    "https://www.google.com/page5",
    "https://college.edu",
    "https://college.edu/page1",
    "https://example.org/page_1",
]


@pytest.fixture
def settings(request: pytest.FixtureRequest) -> Iterable[Settings]:
    exclusions = request.getfixturevalue("exclusions")

    settings = Settings(dict(iter_default_settings()))
    settings.set("SCRACHY_CACHED_RESPONSE_FILTER_EXCLUSIONS", exclusions)
    settings.set("SCRACHY_DB_DIALECT", "sqlite")
    settings.set("SCRACHY_DB_DATABASE", ":memory:")

    # See: https://stackoverflow.com/a/78053441
    settings.delete("TWISTED_REACTOR")

    yield settings


@pytest.fixture
def crawler(request: pytest.FixtureRequest) -> Crawler:
    settings: Settings = request.getfixturevalue("settings")

    crawler = Crawler(MockSpider, settings=settings)

    crawler._apply_settings()

    return crawler


@contextmanager
def populate_cache(engine: SyncAlchemyEngine, crawler: Crawler):
    urls = CACHED_URLS

    assert crawler.request_fingerprinter is not None

    fingerprints = [
        crawler.request_fingerprinter.fingerprint(Request(url)) for url in urls
    ]
    scrape_timestamps = [
        arrow.get(datetime.datetime(2023, 1, 1, 12, i)) for i in range(len(urls))
    ]

    with engine.session_scope() as session:
        repo = SyncResponseRepository(engine)

        for fingerprint, scrape_timestamp, url in zip(
            fingerprints, scrape_timestamps, urls
        ):
            repo.insert(
                Response(
                    fingerprint=fingerprint,
                    scrape_timestamp=scrape_timestamp,
                    url=url,
                    method=True,
                    body=b"",
                    body_length=0,
                ),
                session=session,
            )

        session.flush()

        yield

        session.execute(delete(Response))


@pytest.mark.parametrize(
    "exclusions, raise_expectations",
    [
        (
            [r"google\.com", r"\.edu$"],
            [
                does_not_raise(),
                does_not_raise(),
                does_not_raise(),
                pytest.raises(IgnoreRequest),
                pytest.raises(IgnoreRequest),
                does_not_raise(),
                does_not_raise(),
                does_not_raise(),
                does_not_raise(),
            ],
        ),
        (
            [re.compile(r"\.edu$")],
            [
                pytest.raises(IgnoreRequest),
                pytest.raises(IgnoreRequest),
                does_not_raise(),
                pytest.raises(IgnoreRequest),
                pytest.raises(IgnoreRequest),
                does_not_raise(),
                does_not_raise(),
                does_not_raise(),
                does_not_raise(),
            ],
        ),
    ],
)
def test_cache_filter(
    crawler: Crawler,
    exclusions: list[PatternLike],
    raise_expectations: list,
):
    cache_filter = CachedResponseFilter(crawler)
    spider = crawler.spider

    # TODO Rethink how the fixtures work.
    # This is probably failing because 'populate_cache' and the 'cache_filter'
    # are using different engines and therefore databases.
    with populate_cache(cache_filter._engine, crawler):
        # It's not (easily) possible to test if an item should be skipped due to
        # being stale. That path of code is fairly straight forward and the
        # expiration manager is tested, so it should be correct.
        for url, raise_expectation in zip(
            CACHED_URLS + UN_CACHED_URLS, raise_expectations
        ):
            with raise_expectation:
                cache_filter.process_request(Request(url=url), spider)

        dont_cache = {"dont_cache": True}
        dont_filter = {"dont_filter": True}
        cachable_url = "https://example.org"

        assert (
            cache_filter.process_request(
                Request(url=cachable_url, meta=dont_cache), spider
            )
            is None
        )
        assert (
            cache_filter.process_request(
                Request(url=cachable_url, meta=dont_filter), spider
            )
            is None
        )



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Future Library
from __future__ import annotations

# Standard Library
import dataclasses
import datetime
import logging
import pathlib

from dataclasses import dataclass
from typing import Any, NamedTuple, Optional

# 3rd Party Library
import arrow
import pytest
import yaml

from scrapy.settings import Settings, iter_default_settings

# 1st Party Library
from scrachy.db.models import Response
from scrachy.settings.defaults import filter as filter_settings
from scrachy.settings.defaults import fingerprinter as fingerprinter_settings
from scrachy.settings.defaults import policy as policy_settings
from scrachy.settings.defaults import storage as storage_settings

log = logging.getLogger(__name__)


class TimestampPair(NamedTuple):
    scrape_timestamp: datetime.datetime
    current_timestamp: datetime.datetime


@dataclass(kw_only=True)
class ResponseConfig(yaml.YAMLObject):
    yaml_tag = "!ResponseConfig"
    yaml_loader = yaml.SafeLoader

    scrape_timestamp: datetime.datetime
    fingerprint: str
    url: str
    method: str
    request_body: Optional[bytes] = None
    body: bytes
    meta: Optional[dict[str, Any]] = None
    status: Optional[int] = None
    headers: Optional[str] = None
    extracted_text: Optional[str] = None
    body_length: int
    extracted_text_length: Optional[int] = None

    def to_cached_response(self) -> Response:
        """
        Convert the dataclass to a cached ``Response``.
        """
        kwargs = {f.name: getattr(self, f.name) for f in dataclasses.fields(self)}
        kwargs["fingerprint"] = bytes.fromhex(kwargs["fingerprint"])
        kwargs["scrape_timestamp"] = arrow.get(
            kwargs["scrape_timestamp"].replace(tzinfo=datetime.timezone.utc)
        )

        return Response(**kwargs)


@dataclass(kw_only=True)
class SettingsConfig(yaml.YAMLObject):
    yaml_tag = "!SettingsConfig"
    yaml_loader = yaml.SafeLoader

    settings: dict[str, Any]

    def to_settings(self):
        settings = Settings(dict(iter_default_settings()))
        settings.setmodule(filter_settings)
        settings.setmodule(fingerprinter_settings)
        settings.setmodule(policy_settings)
        settings.setmodule(storage_settings)
        settings.update(self.settings)

        return settings


@pytest.fixture(scope="module")
def example_urls() -> list[str]:
    return [
        "http://www.example.com",
        "https://www.example.com/page1",
        "ftp://www.example.com/page2.html",
        "http://wwwwexample.com",
        "http://w3.example.com",
        "https://anything.edu",
        "http://anything.edu",
        "https://anything.edu",
        "ftp://anything.edu/page1",
        "https://www.anything.edu/page1",
        "foo://test.com",
        "http://test.comhttps://test.com/page1",
    ]


@pytest.fixture(scope="module")
def response_configs() -> list[ResponseConfig]:
    path = pathlib.Path(__file__).with_name("responses.yaml")
    with path.open("r") as fh:
        return yaml.safe_load(fh)


@pytest.fixture(scope="module")
def cached_responses(response_configs: list[ResponseConfig]) -> list[Response]:
    return [r.to_cached_response() for r in response_configs]


@pytest.fixture(scope="module")
def settings_choices() -> dict[str, Settings]:
    path = pathlib.Path(__file__).with_name("settings.yaml")
    with path.open("r") as fh:
        settings = yaml.safe_load(fh)

    return (
        {name: config.to_settings() for name, config in settings.items()}
        if settings
        else {}
    )



#  Copyright 2020 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
The basic data types and classes required to define the SqlAlchemy models.
"""

# Future Library
from __future__ import annotations

# Standard Library
import logging

from typing import Annotated, Any, ClassVar

# 3rd Party Library
from arrow import Arrow
from rwskit.sqlalchemy.base import BaseModel, default_metadata
from sqlalchemy import BigInteger, LargeBinary, MetaData, SmallInteger
from sqlalchemy_utils.types import ArrowType

# 1st Party Library
from scrachy.settings import PROJECT_SETTINGS
from scrachy.utils.sqltypes import ConditionalJson

log = logging.getLogger(__name__)


schema = PROJECT_SETTINGS.get("SCRACHY_DB_SCHEMA")
schema_prefix = f"{schema}." if schema else ""

bigint = Annotated[int, 64]
binary = Annotated[bytes, None]
smallint = Annotated[int, 16]
timestamp = Annotated[Arrow, None]
conditional_json = Annotated[dict[str, Any], None]


BaseModel.registry.update_type_annotation_map({
    bigint: BigInteger,
    binary: LargeBinary,
    conditional_json: ConditionalJson,
    smallint: SmallInteger,
    timestamp: ArrowType,
})


class ScrachyBase(BaseModel):
    __abstract__ = True

    metadata: ClassVar[MetaData] = default_metadata()



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
# Loosely based on: https://hackernoon.com/building-a-to-do-list-app-with-python-data-access-layer-with-sqlalchemy

"""
The Data Access Layer.
"""

# Standard Library
import logging

from typing import Any, Callable, Optional, TypeVar

# 3rd Party Library
from arrow import Arrow
from rwskit.sqlalchemy.engine import AsyncAlchemyEngine, SyncAlchemyEngine
from rwskit.sqlalchemy.repository import AsyncRepository, SyncRepository
from sqlalchemy import Select, select
from sqlalchemy.dialects.postgresql.dml import Insert as PostgresInsert
from sqlalchemy.dialects.sqlite.dml import Insert as SqliteInsert
from sqlalchemy.ext.asyncio import AsyncSession
from sqlalchemy.orm import Session, load_only, selectinload

# 1st Party Library
from scrachy.db.base import ScrachyBase
from scrachy.db.models import Response, ScrapeHistory
from scrachy.settings.defaults.storage import RetrievalMethod

BaseT = TypeVar("BaseT", bound=ScrachyBase)

InsertFunction = Callable[[Any], SqliteInsert | PostgresInsert]


log = logging.getLogger(__name__)


class ResponseRepository:
    def _find_minimal_stmt(self, fingerprint: bytes) -> Select:
        return (
            select(Response)
            .options(load_only(Response.body))
            .where(Response.fingerprint == fingerprint)
        )

    def _find_standard_stmt(self, fingerprint: bytes) -> Select:
        return (
            select(Response)
            .options(
                load_only(
                    Response.body, Response.meta, Response.headers, Response.status
                )
            )
            .where(Response.fingerprint == fingerprint)
        )

    def _find_full_stmt(self, fingerprint: bytes) -> Select:
        return (
            select(Response)
            .options(selectinload(Response.scrape_history))
            .where(
                Response.fingerprint == fingerprint,
            )
        )


class SyncResponseRepository(SyncRepository[Response], ResponseRepository):
    def __init__(self, engine: SyncAlchemyEngine):
        super().__init__(engine, Response)

    def find_timestamp_by_fingerprint(
        self, fingerprint: bytes, session: Optional[Session] = None
    ) -> Optional[Arrow]:
        stmt = select(Response.scrape_timestamp).where(
            Response.fingerprint == fingerprint
        )

        with self.get_or_create_session(session) as local_session:
            return local_session.scalars(stmt).first()

    def find_by_fingerprint(
        self,
        fingerprint: bytes,
        retrieval_method: RetrievalMethod = "full",
        session: Optional[Session] = None,
    ) -> Optional[Response]:
        if retrieval_method == "minimal":
            stmt = self._find_minimal_stmt(fingerprint)
        elif retrieval_method == "standard":
            stmt = self._find_standard_stmt(fingerprint)
        elif retrieval_method == "full":
            stmt = self._find_full_stmt(fingerprint)
        else:
            raise ValueError(f"Unknown retrieval method: {retrieval_method}")

        with self.get_or_create_session(session) as local_session:
            return self._find(stmt, local_session)

    def _find(self, stmt: Select, session: Session) -> Optional[Response]:
        return session.scalars(stmt).one_or_none()


class AsyncResponseRepository(AsyncRepository[Response], ResponseRepository):
    def __init__(self, engine: AsyncAlchemyEngine):
        super().__init__(engine, Response)

    async def find_timestamp_by_fingerprint(
        self, fingerprint: bytes, session: Optional[AsyncSession] = None
    ) -> Optional[Arrow]:
        stmt = select(Response.scrape_timestamp).where(
            Response.fingerprint == fingerprint
        )

        async with self.get_or_create_session(session) as local_session:
            future_result = await local_session.scalars(stmt)
            return future_result.first()

    async def find_by_fingerprint(
        self,
        fingerprint: bytes,
        retrieval_method: RetrievalMethod = "full",
        session: Optional[AsyncSession] = None,
    ) -> Optional[Response]:
        if retrieval_method == "minimal":
            stmt = self._find_minimal_stmt(fingerprint)
        elif retrieval_method == "standard":
            stmt = self._find_standard_stmt(fingerprint)
        elif retrieval_method == "full":
            stmt = self._find_full_stmt(fingerprint)
        else:
            raise ValueError(f"Unknown retrieval method: {retrieval_method}")

        async with self.get_or_create_session(session) as local_session:
            return await self._find(stmt, local_session)

    async def _find(self, stmt: Select, session: AsyncSession) -> Optional[Response]:
        future_result = await session.scalars(stmt)
        return future_result.one_or_none()


class SyncScrapeHistoryRepository(SyncRepository[ScrapeHistory]):
    def __init__(self, engine: SyncAlchemyEngine):
        super().__init__(engine, ScrapeHistory)


class AsyncScrapeHistoryRepository(AsyncRepository[ScrapeHistory]):
    def __init__(self, engine: AsyncAlchemyEngine):
        super().__init__(engine, ScrapeHistory)



#  Copyright 2020 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
"""Middleware for filtering (or ignoring) responses if they are fresh in the cache."""

# Standard Library
import logging

from typing import Optional

# 3rd Party Library
from scrapy.crawler import Crawler
from scrapy.exceptions import IgnoreRequest
from scrapy.http.request import Request
from scrapy.spiders import Spider

# 1st Party Library
from scrachy.db.engine import initialize_engine
from scrachy.db.repositories import SyncResponseRepository
from scrachy.utils.datetime import now_tzaware
from scrachy.utils.request import ExpirationManager
from scrachy.utils.settings import compile_patterns

log = logging.getLogger(__name__)


class CachedResponseFilter:
    @classmethod
    def from_crawler(cls, crawler: Crawler):
        return cls(crawler)

    def __init__(self, crawler: Crawler):
        """
        Sometimes you scrape the same domains multiple times looking for
        new content. However, when crawling them you might encounter pages that
        you have already scraped. If your extraction rules have not changed
        since the last crawl it may not be worth reprocessing those pages.

        This middleware will look to see if a response corresponding to this
        request is already in the cache and is not stale. If the response
        is not in the cache ``process_request`` will return immediately.
        Otherwise, it will use the following rules to determine whether
        the request should be filtered.

        * You can specify a set of patterns to match against the
          request url. Any pattern that matches part of the url will not be
          filtered regardless of whether it is in the cache or
          not. This might be useful after changing parsing rules
          for a set of pages. These are specified using the
          ``SCRACHY_CACHED_RESPONSE_FILTER_EXCLUSIONS`` setting, which takes a
          list of ``re.Patterns`` or strings which can be compiled to regular
          expressions.
        * Setting the request meta key, ``dont_filter`` to ``True``, will
          not be processed by this middleware.
        * Any page that is already excluded from caching via the ``dont_cache``
          request meta key will also never be filtered.

        Any other request that has a fresh response in the cache will be
        filtered.

        :param crawler: The current crawler.
        """
        settings = crawler.spider.settings if crawler.spider else crawler.settings
        self._exclude_patterns = compile_patterns(
            settings.get("SCRACHY_CACHED_RESPONSE_FILTER_EXCLUSIONS")
        )
        self._expiration_manager = ExpirationManager(settings)

        if crawler.request_fingerprinter is None:
            raise ValueError("The request fingerprinter is not set.")
        else:
            self._fingerprinter = crawler.request_fingerprinter

        self._engine = initialize_engine(settings)
        self._response_repository = SyncResponseRepository(self._engine)

    # noinspection PyUnusedLocal
    def process_request(self, request: Request, spider: Optional[Spider]):
        """

        :param request: The Scrapy request.
        :param spider: The Scrapy Spider issuing the request.

        :raises IgnoreRequest: If the item is already cached, and it does
                not meet the requirement to be excluded.
        """
        # If dont_cache or dont_skip is set then don't skip the item.
        if request.meta.get("dont_cache") or request.meta.get("dont_filter"):
            return

        url = request.url
        fingerprint = self._fingerprinter.fingerprint(request)

        # Otherwise check to see if the request is already in the cache and
        # skip further processing if it is.
        scrape_timestamp = self._response_repository.find_timestamp_by_fingerprint(
            fingerprint
        )

        # If the item is not in the cache, then don't filter it
        if scrape_timestamp is None:
            return

        current_timestamp = now_tzaware()

        # If the item in the cache is stale, don't filter it.
        if self._expiration_manager.is_stale(url, scrape_timestamp, current_timestamp):
            return

        # If the url is in the exclusion list then don't filter the request
        # even if it is already in the cache.
        for pattern in self._exclude_patterns:
            if pattern.search(url):
                return

        # Otherwise ignore the cached request
        raise IgnoreRequest()



#  Copyright 2020 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
"""
Additional ``Request`` and ``Response`` classes for working with Selenium
and the ``AlchemyCacheStorage`` backend.

Note: Naming this module ``http`` causes a circular import error, so I've appended
an underscore to avoid conflicts.
"""

# Future Library
from __future__ import annotations

# Standard Library
import logging

from typing import Any, Callable, Optional

# 3rd Party Library
from scrapy.http.request import Request
from scrapy.http.response import Response
from selenium.webdriver.remote.webdriver import WebDriver

log = logging.getLogger(__name__)


DEFAULT_WAIT_TIMEOUT = 30
DEFAULT_POLL_FREQUENCY = 1
DEFAULT_MAX_RETRIES = 10


WaitCondition = Callable[[WebDriver], Any]


ScriptExecutor = Callable[
    [WebDriver, Request], Optional[Response | list[Response] | dict[str, Response]]
]


class SeleniumRequest(Request):
    """
    A subclass of :class:`scrapy.http.Request` that provides extra information for downloading pages using
    Selenium.

    Based off the code from `Scrapy-Selenium <https://github.com/clemfromspace/scrapy-selenium>`_
    """

    def __init__(
        self,
        wait_timeout: float = DEFAULT_WAIT_TIMEOUT,
        wait_until: Optional[WaitCondition] = None,
        poll_frequency: float = DEFAULT_POLL_FREQUENCY,
        screenshot: bool = False,
        script_executor: Optional[ScriptExecutor] = None,
        max_retries: int = DEFAULT_MAX_RETRIES,
        *args,
        **kwargs,
    ):
        """
        A new ``SeleniumRequest``.

        :param wait_timeout: The number of seconds to wait before accessing the data.
        :param wait_until: One of the "selenium.webdriver.support.expected_conditions". The response
                           will be returned until the given condition is fulfilled.
        :param poll_frequency: The sleep interval between calls
        :param screenshot: If ``True``, a screenshot of the page will be taken and the data of the screenshot
                           will be returned in the response "meta" attribute.
        :param script_executor: A function that takes a webdriver and a response as its parameters and optionally
                                returns a list of new response objects as a side effect of its actions (e.g.,
                                executing arbitrary javascript code on the page). Any returned responses will
                                be returned in the ``request.meta`` attribute with the key ``script_result``.
                                Note that the returned responses will not be further processed by any other
                                middleware.
        :param max_retries: The maximum number of times a request will be retried (sent back to
                            scrapy to be rescheduled) if a TimeoutException is raised because the
                            wait_until conditions are not satisfied while downloading the source.

        """
        super().__init__(*args, **kwargs)

        self.wait_timeout = wait_timeout
        self.wait_until = wait_until
        self.poll_frequency = poll_frequency
        self.screenshot = screenshot
        self.script_executor = script_executor
        self.retries = 0
        self.max_retries = max_retries



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""The default settings for configuring the :mod:`Selenium <scrachy.middleware.selenium>`. middleware."""

# Standard Library
from typing import Any, Literal, Optional

WebDriverName = Literal["Chrome", "ChromiumEdge", "Firefox", "Safari"]


SCRACHY_SELENIUM_WEB_DRIVER: WebDriverName = "Chrome"
"""
The name of the webdriver to use.
"""

SCRACHY_SELENIUM_WEB_DRIVER_OPTIONS: list[str] = []
"""
Initialize the webdriver with an ``Options`` object populated with these
options.

For a list of options see:

    * Chrome: `https://peter.sh/experiments/chromium-command-line-switches/`
    * Firefox: `https://www-archive.mozilla.org/docs/command-line-args`
"""

SCRACHY_SELENIUM_WEB_DRIVER_EXTENSIONS: list[str] = []
"""
A list of extensions for the webdriver to load. These should be paths to CRX
files for Chrome or XPI files for Firefox.
"""

SCRACHY_SELENIUM_WEB_DRIVER_PREFERENCES: dict[str, Any] = dict()
"""
A dictionary of preferences for the webdriver to load. For example,
See the **Recognized capabilities** section of
`this page <https://chromedriver.chromium.org/capabilities>`__.
"""

SCRACHY_SELENIUM_LOG_FILE: Optional[str] = None
"""
A file to save logging statements made by a process launched from the
:class:`~scrachy.middleware.selenium.AsyncSeleniumMiddleware`.
"""

SCRACHY_SELENIUM_MAX_RETRIES: int = 5
"""
The maximum number of times to retry if the request fails.
"""

SCRACHY_SELENIUM_PAGE_LOAD_TIMEOUT: float = 10.0
"""
The number of seconds to wait during a call to ``driver.get`` before timing
out if we are unable to get a response from the server.
"""

SCRACHY_SELENIUM_IMPLICIT_WAIT: Optional[float] = None
"""
If not ``None``, set an implicit wait value (``driver.implicitly_wait``) to the
given value.
"""

SCRACHY_SELENIUM_VERIFY_PROXY: bool = False
"""
If ``True``, an exception will be raised when the webdriver is initialized
if requests from the local machine have the same IP address as requests
made from the webdriver.
"""



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
Utilities to initialize and work with the SqlAlchemy engine.
"""

# Standard Library
import logging

from typing import Optional

# 3rd Party Library
from rwskit.sqlalchemy.engine import SyncAlchemyEngine
from scrapy.settings import BaseSettings
from sqlalchemy import create_engine
from sqlalchemy.sql.ddl import CreateSchema

# 1st Party Library
from scrachy.db.base import ScrachyBase
from scrachy.utils.db import construct_url

log = logging.getLogger(__name__)


# Singleton engine for the project. However, it is the responsibility
# of the AlchemyCacheStorage to initialize it on construction.
engine: Optional[SyncAlchemyEngine] = None


def initialize_engine(settings: BaseSettings):
    global engine

    if engine is not None:
        log.debug("The engine is already initialized.")
        return engine  # The engine is already setup

    schema = settings.get("SCRACHY_DB_SCHEMA")
    connect_args = settings.getdict("SCRACHY_DB_CONNECT_ARGS", {})

    # Create the engine
    execution_options = {"schema_translate_map": {None: schema}} if schema else {}
    url = construct_url(settings)

    log.debug(f"Constructing engine from url: {url}")
    sa_engine = create_engine(
        url,
        connect_args=connect_args,
        execution_options=execution_options,
        pool_pre_ping=True,
    )

    log.debug(
        f"Engine initialized from parameters: "
        f"{sa_engine.url.render_as_string(hide_password=True)}."
    )

    # Create the schema if necessary
    if schema is not None:
        with sa_engine.connect() as connection:
            connection.execute(CreateSchema(schema, if_not_exists=True))
            connection.commit()

    # Create the tables if necessary
    ScrachyBase.metadata.create_all(sa_engine)

    engine = SyncAlchemyEngine(sa_engine, ScrachyBase)

    return engine


# def reset_engine():
#     global engine
#     global session_factory

#     if engine is not None:
#         engine.dispose()

#     if session_factory is not None:
#         session_factory.close_all()

#     engine = None
#     session_factory = None


# @contextmanager
# def session_scope():
#     if session_factory is None:
#         raise ValueError("You must initialize the engine first.")

#     session = session_factory()

#     # noinspection PyBroadException
#     try:
#         yield session
#         session.commit()
#     except Exception as e:
#         session.rollback()
#         raise e
#     finally:
#         session.close()



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
#

# Standard Library
import pathlib

from typing import Any, Optional, Protocol

# 3rd Party Library
import pytest

from scrapy import Spider
from scrapy.http.response import Response
from scrapy.settings import Settings, iter_default_settings

# 1st Party Library
from scrachy.settings.defaults import filter as filter_defaults
from scrachy.settings.defaults import selenium as selenium_defaults
from scrachy.settings.defaults.selenium import WebDriverName
from scrachy.settings.defaults.storage import PatternLike


class MockSpider(Spider):
    name = "mock_spider"
    allowed_domains = ["scrapethissite.com"]

    def parse(self, response: Response, **kwargs: Any) -> Any:
        pass


class SettingsFactory(Protocol):
    def __call__(
        self,
        driver_name: Optional[WebDriverName] = None,
        options: Optional[list[str]] = None,
        extensions: Optional[list[str]] = None,
    ) -> Settings: ...


@pytest.fixture
def extensions(request: pytest.FixtureRequest) -> list[str]:
    if not request.param:
        return []

    driver_name = request.node.funcargs["driver"]
    basedir = pathlib.Path(__file__).parent / "selenium" / "extensions"

    if driver_name == "Chrome":
        return [str(f) for f in basedir.glob("*.crx")]
    elif driver_name == "Firefox":
        return [str(f) for f in basedir.glob("*.xpi")]

    raise ValueError(f"Unsupported driver: {driver_name}")


@pytest.fixture
def settings_factory() -> SettingsFactory:
    settings = Settings(dict(iter_default_settings()))
    settings.setmodule(selenium_defaults)
    settings.setmodule(filter_defaults)

    def wrapper(
        driver_name: Optional[WebDriverName] = None,
        options: Optional[list[str]] = None,
        extensions: Optional[list[str]] = None,
        exclusions: Optional[list[PatternLike]] = None,
    ) -> Settings:
        s = settings.copy()
        s["CONCURRENT_REQUESTS"] = 1

        if driver_name:
            s["SCRACHY_SELENIUM_WEB_DRIVER"] = driver_name

        if options:
            s["SCRACHY_SELENIUM_WEB_DRIVER_OPTIONS"] = options

        if extensions:
            s["SCRACHY_SELENIUM_WEB_DRIVER_EXTENSIONS"] = extensions

        s["SCRACHY_CACHED_RESPONSE_FILTER_EXCLUSIONS"] = exclusions or []

        return s

    return wrapper



#  Copyright 2020 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

import unittest

from scrachy.utils.numeric import slice_int, to_int64, to_uint64


class TestUtils(unittest.TestCase):
    def test_slice_int(self):
        # Concatenate two 64bit integers into one 128bit integer and make sure
        # we get back the first integer when slicing the first 64 bits.
        first_int, second_int = 167, 572
        large_int = first_int.to_bytes(8, 'big') + second_int.to_bytes(8, 'big')

        act_int = slice_int(large_int, 64)
        self.assertEqual(first_int, act_int)

        # Make sure we get back the same value we put in if it can be
        # represented by the given number of bits.
        act_int = slice_int(first_int.to_bytes(4, 'big'), 32)
        exp_int = first_int

        self.assertEqual(act_int, exp_int)

        # Most of the bytes at the front of this integer are 0, so slicing
        # only the first 32 bits will return 0.
        act_int = slice_int(first_int.to_bytes(8, 'big'), 32)
        exp_int = 0

        self.assertEqual(act_int, exp_int)

    def test_signed_to_unsigned(self):
        max_unsigned = 18446744073709551615

        act_signed = to_int64(max_unsigned)
        exp_signed = -1

        self.assertEqual(act_signed, exp_signed)

        act_unsigned = to_uint64(exp_signed)
        exp_unsigned = max_unsigned

        self.assertEqual(act_unsigned, exp_unsigned)


if __name__ == '__main__':
    unittest.main()



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.


# Standard Library
import datetime
import logging

# 3rd Party Library
import arrow
import pytest

from scrapy.settings import Settings

# 1st Party Library
from scrachy.middleware.httpcache import ExpirationManager
from tests.utils import parse_date

log = logging.getLogger("test_expiration_manager")


@pytest.fixture
def settings(
    settings_choices: dict[str, Settings], request: pytest.FixtureRequest
) -> Settings:
    return settings_choices[request.param]  # noqa


@pytest.fixture
def scrape_timestamp(request: pytest.FixtureRequest) -> datetime.datetime:
    return parse_date(request.param).replace(tzinfo=datetime.timezone.utc)  # noqa


@pytest.fixture
def current_timestamp(request: pytest.FixtureRequest) -> datetime.datetime:
    return parse_date(request.param).replace(tzinfo=datetime.timezone.utc)  # noqa


@pytest.mark.parametrize(
    "settings, url, scrape_timestamp, current_timestamp, expected_staleness",
    [
        (
            "defaults",
            "http://www.example.com",
            "2020-01-01T12:00",
            "2025-01-01T12:00",
            False,
        ),
        (
            "defaults",
            "ftp://anything.edu",
            "2020-01-01T12:00",
            "2025-01-01T12:00",
            False,
        ),
        (
            "defaults_with_expiration_secs",
            "http://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:02",
            False,
        ),
        (
            "defaults_with_expiration_secs",
            "http://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:03",
            True,
        ),
        (
            "defaults_with_activation_secs",
            "http://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:02",
            True,
        ),
        (
            "defaults_with_activation_secs",
            "http://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:03",
            False,
        ),
        (
            "defaults_with_expiration_schedule",
            "http://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T13:00",
            False,
        ),
        (
            "defaults_with_expiration_schedule",
            "http://www.example.com",
            "2020-01-01T12:01",
            "2020-01-02T13:01",
            True,
        ),
        (
            "defaults_with_activation_pat",
            "http://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:02",
            True,
        ),
        (
            "defaults_with_activation_pat",
            "https://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:02",
            True,
        ),
        (
            "defaults_with_activation_pat",
            "https://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:03",
            False,
        ),
        (
            "defaults_with_activation_pat",
            "ftp://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:02",
            False,
        ),
        (
            "defaults_with_expiration_pat",
            "http://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:02",
            False,
        ),
        (
            "defaults_with_expiration_pat",
            "https://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:02",
            False,
        ),
        (
            "defaults_with_expiration_pat",
            "https://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:03",
            True,
        ),
        (
            "defaults_with_expiration_pat",
            "ftp://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:02",
            False,
        ),
        (
            "defaults_with_expiration_schedule_pat",
            "http://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T13:01",
            False,
        ),
        (
            "defaults_with_expiration_schedule_pat",
            "https://www.example.com",
            "2020-01-01T12:01",
            "2020-01-02T12:01",
            True,
        ),
        (
            "defaults_with_expiration_schedule_pat",
            "ftp://www.example.com",
            "2020-01-01T12:01",
            "2020-01-02T12:01",
            False,
        ),
        (
            "all_expiration_values",
            "http://test.com",
            "2020-01-01T12:01",
            "2020-01-02T12:01:05",
            True,
        ),  # Non matching
        (
            "all_expiration_values",
            "http://test.com",
            "2020-01-01T12:01",
            "2020-01-01T12:01:20",
            False,
        ),  # Non matching
        (
            "all_expiration_values",
            "http://test.com",
            "2020-01-01T12:01",
            "2020-01-01T12:18:00",
            True,
        ),  # Non matching
        # This has been in the cache long enough to be active, but not long enough to exceed expiration_secs,
        # but it should expire due to the default schedule.
        (
            "all_expiration_values",
            "http://test.com",
            "2020-01-29T23:30",
            "2020-01-31T00:00:00",
            True,
        ),  # Non matching
        (
            "all_expiration_values",
            "http://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:00:40",
            True,
        ),
        (
            "all_expiration_values",
            "http://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:03:00",
            False,
        ),
        (
            "all_expiration_values",
            "http://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:10:00",
            True,
        ),
        (
            "all_expiration_values",
            "http://www.example.com",
            "2020-01-01T11:55",
            "2020-01-01T12:00:01",
            True,
        ),
        (
            "all_expiration_values",
            "https://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T12:02:40",
            True,
        ),
        (
            "all_expiration_values",
            "https://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T13:10:00",
            False,
        ),
        (
            "all_expiration_values",
            "https://www.example.com",
            "2020-01-01T12:01",
            "2020-01-01T14:00:00",
            True,
        ),
        (
            "all_expiration_values",
            "https://www.example.com",
            "2020-01-01T11:50",
            "2020-01-01T12:00:01",
            False,
        ),
        (
            "all_expiration_values",
            "ftp://www.example.com",
            "2020-01-01T11:59",
            "2020-01-01T12:00:01",
            True,
        ),
    ],
    indirect=["settings", "scrape_timestamp", "current_timestamp"],
)
def test_manager(
    settings: Settings,
    url: str,
    scrape_timestamp: datetime.datetime,
    current_timestamp: datetime.datetime,
    expected_staleness: bool,
):
    manager = ExpirationManager(settings)

    log.debug(f"scrape_timetamp: {type(scrape_timestamp)}")

    assert (
        manager.is_stale(url, arrow.get(scrape_timestamp), arrow.get(current_timestamp))
        == expected_staleness
    )



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.
#

# Standard Library
import logging
import time

from pathlib import Path
from typing import Optional, Protocol

# 3rd Party Library
import pytest

from scrapy import Spider
from scrapy.crawler import Crawler
from scrapy.http.request import Request
from scrapy.http.response.html import HtmlResponse
from scrapy.settings import Settings
from scrapy.utils.python import to_bytes
from sqlalchemy import delete
from w3lib.http import headers_raw_to_dict

# 1st Party Library
from scrachy.db.models import Response, ScrapeHistory
from scrachy.middleware.httpcache import AlchemyCacheStorage
from scrachy.settings.defaults.storage import RetrievalMethod
from tests.utils import is_postgresql_setup

log = logging.getLogger("test_storage")


DATABASE_DIALECTS = ["sqlite", "postgresql"] if is_postgresql_setup() else ["sqlite"]


class SettingsFactory(Protocol):
    def __call__(
        self, hasher: Optional[str], extractor: Optional[str], method: RetrievalMethod
    ) -> Settings: ...


@pytest.fixture
def headers(request: pytest.FixtureRequest) -> Optional[str]:
    if request.param:  # noqa
        return """
            Host: code.tutsplus.com
            User-Agent: Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US; rv:1.9.1.5) Gecko/20091102 Firefox/3.5.5 (.NET CLR 3.5.30729)
            Accept: text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8
            Accept-Language: en-us,en;q=0.5
            Accept-Encoding: gzip,deflate
            Accept-Charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7
            Keep-Alive: 300
            Connection: keep-alive
            Cookie: PHPSESSID=r2t5uvjq435r4q7ib3vtdjq120
            Pragma: no-cache
            Cache-Control: no-cache
        """

    return None


@pytest.fixture
def request_body(request: pytest.FixtureRequest) -> Optional[str]:
    if request.param:  # noqa
        return """
            POST /test HTTP/1.1
            Host: foo.example
            Content-Type: application/x-www-form-urlencoded
            Content-Length: 27
            field1=value1&field2=value2
        """

    return None


@pytest.fixture
def response_body(request: pytest.FixtureRequest) -> Optional[str]:
    if request.param == "page_1":
        return '<!-- http://help.websiteos.com/websiteos/example_of_a_simple_html_page.htm -->\n<HTML>\n<HEAD>\n  <TITLE>Your Title Here</TITLE>\n</HEAD>\n<BODY BGCOLOR="FFFFFF">\n<CENTER><IMG SRC="clouds.jpg" ALIGN="BOTTOM"></CENTER>\n<HR>\n<a href="http://somegreatsite.com">Link Name</a>\nis a link to another nifty site\n<H1>This is a Header</H1>\n<H2>This is a Medium Header</H2>\nSend me mail at <a href="mailto:support@yourcompany.com">support@yourcompany.com</a>.\n<P> This is a new paragraph!</P>\n<P><B>This is a new paragraph!</B>\n  <BR> <B><I>This is a new sentence without a paragraph break, in bold italics.</I></B>\n</P>\n<HR>\n</BODY>\n</HTML>'


@pytest.fixture
def extracted_text(request: pytest.FixtureRequest) -> Optional[str]:
    if request.param == ("page_1", "html.parser"):  # noqa
        return "Your Title Here\nLink Name\nis a link to another nifty site\nThis is a Header\nThis is a Medium Header\nSend me mail at\nsupport@yourcompany.com\n.\nThis is a new paragraph!\nThis is a new paragraph!\nThis is a new sentence without a paragraph break, in bold italics."
    elif request.param == ("page_1", "ArticleExtractor"):  # noqa
        return "Link Name is a link to another nifty site\nThis is a Header\nThis is a Medium Header\nSend me mail at support@yourcompany.com .\nThis is a new paragraph!\nThis is a new paragraph!\nThis is a new sentence without a paragraph break, in bold italics."

    return None


@pytest.fixture
def settings_factory(settings_choices: dict[str, Settings], tmp_path: Path):
    def wrapper(
        hasher: Optional[str],
        extractor: Optional[str],
        retrieval_method: RetrievalMethod = "full",
    ) -> Settings:
        settings = settings_choices["defaults"].copy()

        db_path = tmp_path / "test_storage.db"
        db_path.unlink(missing_ok=True)

        settings.set("SCRACHY_DB_DIALECT", "sqlite")
        settings.set("SCRACHY_DB_DATABASE", str(db_path))

        if hasher is not None:
            finger_class = "scrachy.utils.request.DynamicHashRequestFingerprinter"
            settings.set("REQUEST_FINGERPRINTER_CLASS", finger_class)
            settings.set("SCRACHY_REQUEST_FINGERPRINTER_HASHER_CLASS", hasher)

        if extractor is not None:
            if extractor in ("html.parser", "lxml", "html5lib"):
                settings.set(
                    "SCRACHY_CONTENT_EXTRACTOR",
                    "scrachy.content.bs4.BeautifulSoupExtractor",
                )
                settings.set("SCRACHY_CONTENT_BS4_PARSER", extractor)
            elif extractor in ("DefaultExtractor", "ArticleExtractor"):
                settings.set(
                    "SCRACHY_CONTENT_EXTRACTOR",
                    "scrachy.content.boilerpipe.BoilerpipeExtractor",
                )
                settings.set(
                    "SCRACHY_BOILERPY_EXTRACTOR", f"boilerpy3.extractors.{extractor}"
                )

        settings.set("SCRACHY_CACHE_RESPONSE_RETRIEVAL_METHOD", retrieval_method)
        settings.set("HTTPCACHE_EXPIRATION_SECS", 0.5)

        # See: https://stackoverflow.com/a/78053441
        settings.delete("TWISTED_REACTOR")

        return settings

    return wrapper


class MockSpider(Spider):
    def __init__(self, name: str, **kwargs):
        super().__init__(name, **kwargs)


@pytest.mark.parametrize(
    "url, hasher, extractor, method, headers, request_body, response_body, extracted_text, wait",
    [
        (
            "https://www.example.com",
            None,
            None,
            "minimal",
            False,
            False,
            "page_1",
            None,
            0.0,
        ),
        (
            "https://www.example.com",
            None,
            "html.parser",
            "full",
            False,
            False,
            "page_1",
            ("page_1", "html.parser"),
            0.0,
        ),
        (
            "https://www.example.com",
            None,
            "ArticleExtractor",
            "full",
            False,
            False,
            "page_1",
            ("page_1", "ArticleExtractor"),
            0.0,
        ),
        (
            "https://www.example.com",
            "hashlib.md5",
            "ArticleExtractor",
            "full",
            True,
            True,
            "page_1",
            ("page_1", "ArticleExtractor"),
            1.0,
        ),
    ],
    indirect=["headers", "request_body", "response_body", "extracted_text"],
)
def test_storage(
    settings_factory: SettingsFactory,
    url: str,
    hasher: str,
    extractor: str,
    method: RetrievalMethod,
    headers: Optional[str],
    request_body: Optional[str],
    response_body: str,
    extracted_text: Optional[str],
    wait: float,
):
    settings = settings_factory(hasher, extractor, method)
    crawler = Crawler(MockSpider, settings)
    crawler._apply_settings()
    spider = MockSpider.from_crawler(crawler, name="mock_spider")

    storage = AlchemyCacheStorage(settings)
    storage.open_spider(spider)

    # The database is persistent across calls, so we need to clear it each
    # time.
    with storage.engine.session_scope() as session:
        session.execute(delete(Response))
        session.execute(delete(ScrapeHistory))

    request = Request(
        url=url,
        body=request_body,
        headers=headers_raw_to_dict(to_bytes(headers)) if headers else None,
        method="POST" if request_body else "GET",
    )

    response = HtmlResponse(url=url, body=response_body, encoding="utf-8", status=200)

    storage.store_response(spider, request, response)

    time.sleep(wait)

    act_response = storage.retrieve_response(spider, request)

    log.debug(f"wait: {wait} method: {method} act_response: {act_response}")
    if wait > 0.5:
        assert act_response is None
    else:
        assert act_response is not None
        assert act_response.url == request.url == response.url
        assert act_response.text == response.text
        assert act_response.meta.get("extracted_text") == extracted_text

    storage.close_spider(spider)



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""The default settings for configuring :class:`~scrachy.middleware.httpcache.AlchemyCacheStorage`."""

from __future__ import annotations

# Standard Library
import re

from typing import TYPE_CHECKING, Any, Literal, Optional, Type

# 3rd Party Library
from cron_converter import Cron

# 1st Party Library
from scrachy.content import ContentExtractor

if TYPE_CHECKING:
    from boilerpy3.extractors import Extractor

BeautifulSoupParser = Literal["html.parser", "lxml", "lxml-xml", "html5lib"]
RetrievalMethod = Literal["minimal", "standard", "full"]
PatternLike = str | re.Pattern
Schedulable = str | Cron

# Expiration ##################################################################
SCRACHY_CACHE_ACTIVATION_SECS: float = 0
"""
Consider any page that is in the cache stale (do not retrieve it) if it has
not been in the cache for at least this many seconds. This might be used
for sites that initially post unreliable or partial data then update it
with better data after some period of time but then rarely change it again.
"""

SCRACHY_CACHE_ACTIVATION_SECS_PATTERNS: list[tuple[PatternLike, float]] = []
"""
A list of tuples consisting of a pattern and a delay time in seconds. The
pattern should either be a :class:`re.Pattern` or a string that can be
compiled to one. Any url that matches this pattern will use the value in
the second element of the tuple as its activation delay.

See: :const:`SCRACHY_CACHE_ACTIVATION_SECS`.
"""

SCRACHY_CACHE_EXPIRATION_SECS_PATTERNS: list[tuple[PatternLike, float]] = []
"""
Similar to :const:`SCRACHY_CACHE_ACTIVATION_SECS_PATTERNS`, but overrides
``HTTPCACHE_EXPIRATION_SECS`` for matching urls.
"""

SCRACHY_CACHE_EXPIRATION_SCHEDULE: Optional[Schedulable] = None
"""
Expire all responses that do not match a
:const:`schedule pattern <SCRACHY_CACHE_EXPIRATION_SCHEDULE_PATTERNS>` in the
cache  according to this schedule.
"""

SCRACHY_CACHE_EXPIRATION_SCHEDULE_PATTERNS: Optional[
    list[tuple[PatternLike, Schedulable]]
] = []
"""
Expire any response who's URL matches the given pattern according to the
corresponding schedule.
"""

# Encoding ####################################################################
SCRACHY_CACHE_DEFAULT_ENCODING: str = "utf-8"
"""
Sometimes it is not possible to determine the encoding of a page because it was
not set properly at the source. But this also seems to happen for compressed
pages which have an encoding based on the compression algorithm (e.g., gzip).
However, Scrapy will raise an exception when constructing a
:class:`scrapy.http.TextResponse` if it can't determine the encoding.
To avoid these issues you can specify a default encoding to use when Scrapy
fails to automatically identify a compatible one.
"""

# Retrieval ###################################################################
SCRACHY_CACHE_RESPONSE_RETRIEVAL_METHOD: RetrievalMethod = "standard"
"""
The cache stores quite a bit of information about each response. Not all of this
information is useful for a given scraping task or might only be used for post
scraping analysis. To help avoid loading unnecessary information you can select
one of three retrieval methods that vary in the amount of data they retrieve.
All three methods return some subclass of
:class:`~scrapy.http.TextResponse` object,  but may have ``null`` values for
some of the properties.
"""

# Database Settings ###########################################################
SCRACHY_DB_DIALECT: str = "sqlite"
"""
This specifies the database dialect to use and must be supported by
`SQLAlchemy <https://docs.sqlalchemy.org/en/20/dialects/>`_
"""

SCRACHY_DB_DRIVER: Optional[str] = None
"""
This specifies the name of the driver used to connect to the database. It must
be a name recognized by
`SQLAlchemy <https://docs.sqlalchemy.org/en/13/core/engines.html#supported-databases>`_
or ``None`` to use the default driver. Note, the selected driver (including the
default) must be installed separately prior to using it.
"""

SCRACHY_DB_HOST: Optional[str] = None
"""
The hostname (or ip address) where the database server is running. This should
be ``None`` for sqlite databases. For other databases, the hostname is assumed
to be ``localhost`` if this setting is ``None``.
"""

SCRACHY_DB_PORT: Optional[int] = None
"""
The port number the database server is listening on. This should be ``None``
for sqlite databases. For other databases, the default port for the database
server is used when this setting is ``None``.
"""

SCRACHY_DB_DATABASE: Optional[str] = None
"""
For sqlite this is the path to the database file and it will be created if it
does not already exist. For other dialects this is the name of the database
where the cached items will be stored. The database must exist prior to running
any crawlers, but the backend will create all necessary tables. This requires
that the database user have sufficient privileges to do so. If the value is
``None`` for the sqlite dialect, an in memory database will be used (which is
probably not what you want). For all other dialects ``None`` is not permitted.
"""

SCRACHY_DB_SCHEMA: Optional[str] = None
"""
This will set the schema for databases that support them (e.g., PostgreSQL).
"""

SCRACHY_DB_USERNAME: Optional[str] = None
"""
The username used to connect to the database.
"""

SCRACHY_DB_PASSWORD: Optional[str] = None
"""
The password (if any) used to connect to the database. It is not recommended to
store this directly in the settings file. Instead, it should be loaded
dynamically, e.g., using environment variables or ``python-dotenv``.
"""

SCRACHY_DB_CONNECT_ARGS: dict[str, Any] = dict()
"""
Any other arguments that should be passed to :func:`sqla.create_engine`. For
example, you could use the following ``dict`` to connect to postgresql using
ssl:

.. code-block::

    {
        sslrootcert: "path.to.rootcert",
        sslcert: "path.to.clientcert",
        sslkey: "path.to.clientkey",
        sslmode: "verify-full"
    }
"""

# History #####################################################################
SCRACHY_CACHE_SAVE_HISTORY: bool = False
"""
Whether or not to store the full scrape history for each page (identified by
its fingerprint).
"""

# Content Extraction Settings #################################################
SCRACHY_CONTENT_EXTRACTOR: Optional[str | Type[ContentExtractor]] = None
"""
A class implementing the :class:`~scrachy.content.ContentExtractor` protocol
or an import path to a class implementing it. Scrachy provides two
implementations.

    * :class:`scrachy.content.bs4.BeautifulSoupExtractor`
    * :class:`scrachy.content.boilerpipe.BoilerpipeExtractor`
"""

SCRACHY_CONTENT_BS4_PARSER: Optional[BeautifulSoupParser] = "html.parser"
"""
The
`parser <https://www.crummy.com/software/BeautifulSoup/bs4/doc/#installing-a-parser>`_
to use for constructing the DOM. It can be one of the following: [html.parser,
lxml, lxml-xml, html5lib]. By default, it will use ``html.parser``, but lxml
or html5lib are probably preferred.
"""

SCRACHY_BOILERPY_EXTRACTOR: Optional[str | Type[Extractor]] = (
    "boilerpy3.extractors.DefaultExtractor"
)
"""
A boilerpy ``Extractor`` class or the import path to one of the classes. See:
`the usage section <https://github.com/jmriebold/BoilerPy3#extractors>`_ of
the boilerpy3 documentation.
"""

# Simhash is no longer supported



"""A ``Settings`` class that assumes all setting names start with a given ``prefix``."""

# Standard Library
from typing import Any, Mapping, Optional, Union

# 3rd Party Library
from scrapy.settings import Settings, _SettingsKeyT

_SettingsInputT = Union[Mapping[_SettingsKeyT, Any], str, None]


class PrefixedSettings(Settings):
    """A ``Settings`` class that assumes all setting names start with a given ``prefix``."""

    def __init__(
        self,
        prefix: str,
        values: _SettingsInputT = None,
        priority: int | str = "project",
    ):
        super().__init__(values, priority)

        self._prefix = prefix

    def full_key(self, name: _SettingsKeyT) -> _SettingsKeyT:
        """Get the full key for the setting, including the prefix, from the given ``name``.

        If the ``name`` is a string, the full key is constructed from the
        following pattern ``{prefix}_{name}``. Otherwise, the ``name`` is
        returned as is.

        Parameters
        ----------
        name : _SettingsKeyT
            The (unprefixed) name of the setting.

        Returns
        -------
        _SettingsKeyT
            The full name of the setting (or unaltered key if it is not a string.)
        """
        if isinstance(name, str):
            return f"{self._prefix}_{name}"

        return name

    def get(self, name: _SettingsKeyT, default: Any = None) -> Any:
        return super().get(self.full_key(name), default)

    def getbool(self, name: _SettingsKeyT, default: bool = False) -> bool:
        return super().getbool(self.full_key(name), default)

    def getint(self, name: _SettingsKeyT, default: int = 0) -> int:
        return super().getint(self.full_key(name), default)

    def getfloat(self, name: _SettingsKeyT, default: float = 0) -> float:
        return super().getfloat(self.full_key(name), default)

    def getlist(
        self, name: _SettingsKeyT, default: Optional[list[Any]] = None
    ) -> list[Any]:
        return super().getlist(self.full_key(name), default)

    def getdict(
        self, name: _SettingsKeyT, default: Optional[dict[Any, Any]] = None
    ) -> dict[Any, Any]:
        return super().getdict(self.full_key(name), default)

    def getdictorlist(
        self,
        name: _SettingsKeyT,
        default: dict[Any, Any] | list[Any] | tuple[Any] | None = None,
    ) -> dict[Any, Any] | list[Any]:
        return super().getdictorlist(self.full_key(name), default)
        return super().getdictorlist(self.full_key(name), default)



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""The default settings for configuring :class:`~scrachy.middleware.httpcache.BlacklistPolicy`."""

# Standard Library
import re

from typing import Callable

SCRACHY_POLICY_BASE_CLASS: str | Callable = "scrapy.extensions.httpcache.DummyPolicy"
"""
The base policy the :class:`~scrachy.middleware.httpcache.BlacklistPolicy`
will wrap around. The policy can be specified as the full import path to
the class or a class object itself. Either way the class constructor must
accept a :class:`~scrapy.settings.Settings` object as its first parameter.
"""

SCRACHY_POLICY_EXCLUDE_URL_PATTERNS: list[str | re.Pattern] = []
"""
A list of regular expression patterns, as compilable strings or
:class:`re.Pattern` objects. Any url matching any of these patterns will be
excluded from being processed by the
:class:`~scrapy.downloadermiddlewares.httpcache.HttpCacheMiddleWare`.
"""



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""Classes and utilities for extracting textual content from the HTML body."""

# Standard Library
from typing import Protocol

# 3rd Party Library
from scrapy.settings import Settings


class ContentExtractor(Protocol):
    def get_content(self, html: str) -> str:
        """Get the desired textual content from the HTML.

        Parameters
        ----------
        html : str
            The textual HTML to process.

        Returns
        -------
        str
            The desired content (e.g., text with tags removed).
        """
        ...


class BaseContentExtractor(ContentExtractor):
    def __init__(self, settings: Settings):
        """A content extractor base class that keeps track of the project middleware.

        Parameters
        ----------
        settings : Settings
            The Scrapy ``Settings``.
        """
        super().__init__()

        self.settings = settings



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
Some utility classes for sending messages between the
:class:`~scrachy.middleware.selenium.AsyncSeleniumMiddleware` and the
:mod:`~scrachy.cli.webdriver_server`. It also includes the primary
functionality for processing requests with Selenium. Each Selenium middleware
is a thin wrapper around these functions.
"""

# Future Library
from __future__ import annotations

# Standard Library
import logging
import threading
import time

from typing import Any, Optional, Type, cast

# 3rd Party Library
from scrapy.exceptions import IgnoreRequest
from scrapy.http.request import Request
from scrapy.http.response.html import HtmlResponse
from scrapy.utils.misc import load_object
from selenium import webdriver
from selenium.common import TimeoutException, WebDriverException
from selenium.webdriver import Proxy
from selenium.webdriver.chromium.options import ChromiumOptions
from selenium.webdriver.common.by import By
from selenium.webdriver.common.options import ArgOptions
from selenium.webdriver.common.proxy import ProxyType
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support.wait import WebDriverWait

# 1st Party Library
from scrachy.http_ import SeleniumRequest
from scrachy.settings.defaults.selenium import WebDriverName

log = logging.getLogger(__name__)


webdriver_import_base = "selenium.webdriver"


class BufferIncompleteError(Exception):
    """Raised when the amount of data received is less than expected."""

    def __init__(self, *args):
        super().__init__(*args)


class ShutdownRequest:
    """A message token indicates the webdriver server should terminate."""

    def __int__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class UnknownMessageType:
    """
    A message token indicates the webdriver server received an unknown message
    type.
    """

    def __init__(self, message_type: str, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.message_type = message_type


class UnhandledError:
    """
    A message token indicates that an unhandled exception was raised during the
    processing of a ``SeleniumRequest``.
    """

    def __init__(self, exception: Exception, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.exception = exception


def initialize_driver(
    driver_name: WebDriverName,
    options: list[str],
    extensions: list[str],
    preferences: Optional[dict[str, Any]] = None,
    page_load_timeout: Optional[float] = None,
    implicit_wait: Optional[float] = None,
    verify_proxy: bool = False,
) -> WebDriver:
    """
    Construct a new webdriver instance with the given options and extensions.

    :param page_load_timeout:
    :param preferences:
    :param driver_name: The class name of the webdriver to construct.
    :param options: The list of options for the webdriver to use.
    :param extensions: The list of extensions to install.
    :param implicit_wait: The amount of time to implicitly wait.
    :param verify_proxy: If ``True`` check that the IP address of a request
           from the host machine is not the same as a request from the
           webdriver (if using a proxy).

    :return: The webdriver.
    """
    driver_cls: Type[WebDriver] = load_object(f"{webdriver_import_base}.{driver_name}")
    driver_options: ArgOptions = load_object(
        f"{webdriver_import_base}.{driver_name}Options"
    )()

    log.debug(f"Initializing {driver_name} driver.")

    for option in options:
        if option.startswith("--proxy-server=") and driver_name == "Firefox":
            # Note the httpProxy and sslProxy must both be set even if the proxy only
            # uses http.
            proxy_server = option.replace("--proxy-server=", "")
            proxy = {
                "proxyType": ProxyType.MANUAL,
                "httpProxy": proxy_server,
                "sslProxy": proxy_server,
            }

            driver_options.proxy = Proxy(proxy)
        elif not option.startswith("--class="):
            log.debug(f"Adding option '{option}' to the WebDriver.")
            driver_options.add_argument(option)
        else:
            log.warning(
                "The --class option is reserved for use by Scrachy to help clean up "
                "rogue processes that are not terminated by `driver.quit()`."
            )

    if preferences and driver_name == "Firefox":
        for preference_name, value in preferences.items():
            driver_options.set_preference(preference_name, value)  # type: ignore
    else:
        driver_options.add_experimental_option("prefs", preferences)  # type: ignore

    if "chrom" in driver_name.lower():
        driver_options.add_argument("--class=selenium")

    # Chrome loads the extensions from the options
    if driver_name == "Chrome" and extensions:
        driver_options = cast(ChromiumOptions, driver_options)
        for extension in extensions:
            driver_options.add_extension(extension)  # noqa

    driver = driver_cls(options=driver_options)

    # Firefox appears to load the extension directly from the driver
    if driver_name == "Firefox" and extensions:
        for extension in extensions:
            driver = cast(webdriver.Firefox, driver)

            try:
                driver.install_addon(extension, temporary=True)
            except FileNotFoundError:
                log.error(f"Addon '{extension}' not found. Skipping.")

    # implicitly_wait applies to all find element calls.
    # set_page_load_timeout applies to 'get' and 'navigate' calls.
    if page_load_timeout is not None:
        driver.set_page_load_timeout(page_load_timeout)

    if implicit_wait is not None:
        driver.implicitly_wait(implicit_wait)

    if verify_proxy:
        my_ip = _get_local_ip()

        retry_attempts = 0
        while True:
            try:
                driver_ip = _get_driver_ip(driver)
            except Exception as e:
                retry_attempts += 1
                log.error(
                    f"Failed to get the driver IP address (retry {retry_attempts}): {e}"
                )
                time.sleep(5)
            else:
                break

        log.debug(f"local_ip: {my_ip}, proxy_ip: {driver_ip}")
        if my_ip == driver_ip:
            driver.quit()
            raise ValueError(
                "The local IP address is the same as the driver IP address"
            )

    return driver


def process_request(
    driver: WebDriver, request: Request
) -> Optional[HtmlResponse | Request]:
    if not isinstance(request, SeleniumRequest):
        # Let some other downloader handle this request
        return None

    request = cast(SeleniumRequest, request)

    def _load_page(driver_: WebDriver, url_: str):
        driver_.get(url_)

    try:
        thread = threading.Thread(target=_load_page, args=(driver, request.url))
        thread.start()
        thread.join(timeout=request.wait_timeout)

        if thread.is_alive():
            log.error(
                f"The page for request '{request.url}' did not load properly within "
                f"{request.wait_timeout} seconds. The most likely cause is a misbehaving proxy. "
                f"Terminating the webdriver."
            )
            raise TimeoutException()
        # driver.get(request.url)
    except TimeoutException:
        raise WebDriverException(
            "Failed to get the request because of a page load timeout."
        )
        # request.retries += 1
        # if request.retries > request.max_retries:
        #     log.error(
        #         f"The request for {request.url} reached the maximum number of retries and "
        #         f"could not be processed."
        #     )
        #     raise IgnoreRequest()
        #
        #
        #
        #
        # # Don't increase the wait time on each retry
        # page_load_timeout = driver.timeouts.page_load
        # sleep_time = page_load_timeout
        # log.warning(
        #     f"The request {request} reached the page_load_timeout of {page_load_timeout}s. "
        #     f"Sleeping for {sleep_time}s and then rescheduling the request "
        #     f"({request.retries} out of {request.max_retries})."
        # )
        #
        # time.sleep(sleep_time)
        #
        # request.dont_filter = True
        # return request
    except Exception as e:
        log.error(f"An unknown exception has occurred: {e}")
        raise WebDriverException(f"An unknown exception has occurred: {e}")
    else:
        set_cookies(driver, request)

    try:
        wait_for_page(driver, request)
    except TimeoutException:
        # The timeout could actually be from a page timeout (i.e., from the
        # call to `driver.get`) or from a WebDriverWait condition (i.e.,
        # a call to `wait_for_page`). It doesn't really matter, but log
        # message will be incorrect for a page timeout.
        request.retries += 1
        if request.retries > request.max_retries:
            log.error(
                f"The request for {request.url} reached the maximum number of retries and "
                f"could not be processed."
            )
            raise IgnoreRequest()

        # Sleep longer and longer each time we time out
        sleep_time = request.wait_timeout * request.retries

        log.warning(
            f"The page for request {request} did not load properly after {request.wait_timeout} "
            f"seconds. Sleeping for {sleep_time}s and then rescheduling the request "
            f"({request.retries} out of {request.max_retries})."
        )

        time.sleep(sleep_time)

        request.dont_filter = True
        return request

    take_screenshot(driver, request)

    execute_script(driver, request)

    response = make_response(driver, request)

    return response


def set_cookies(driver: WebDriver, request: SeleniumRequest):
    for cookie_name, cookie_value in request.cookies.items():  # type: ignore
        driver.add_cookie({"name": cookie_name, "value": cookie_value})


def wait_for_page(driver: WebDriver, request: SeleniumRequest):
    if request.wait_until:
        try:
            WebDriverWait(
                driver, request.wait_timeout, poll_frequency=request.poll_frequency
            ).until(request.wait_until)
        except TimeoutException as e:
            log.error(f"Wait condition timed out for url: '{request.url}'")
            raise e


def take_screenshot(driver: WebDriver, request: SeleniumRequest):
    if request.screenshot:
        request.meta["screenshot"] = driver.get_screenshot_as_png()


def make_response(driver: WebDriver, request: SeleniumRequest) -> HtmlResponse:
    return HtmlResponse(
        url=driver.current_url,
        body=driver.page_source,
        encoding="utf-8",
        request=request,
        flags=["selenium"],
    )


def execute_script(driver: WebDriver, request: SeleniumRequest):
    if request.script_executor is not None:
        request.meta["script_result"] = request.script_executor(driver, request)


def _get_local_ip() -> str:
    # 3rd Party Library
    import requests

    response = requests.get("https://api.ipify.org?format=json")
    return response.json()["ip"]


def _get_driver_ip(driver: WebDriver) -> str:
    driver.get("https://api.ipify.org")
    data = driver.find_element(By.CSS_SELECTOR, "body")
    return data.text.strip()



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Standard Library
import logging

from contextlib import contextmanager
from typing import Any, Generator, Optional

# 3rd Party Library
from psycopg import Connection
from rwskit.sqlalchemy.config import DatabaseConnectionConfig
from sqlalchemy import Engine

# 1st Party Library
from scrachy.db.base import ScrachyBase

log = logging.getLogger(__name__)


@contextmanager
def sync_engine_manager(
    dialect: str, postgresql: Optional[Connection] = None
) -> Generator[Engine, Any, None]:
    if dialect == "postgresql" and postgresql is not None:
        info = postgresql.info
        config = DatabaseConnectionConfig(
            drivername="postgresql+psycopg",
            username=info.user,
            host=info.host,
            port=info.port,
            database=info.dbname,
        )
        engine = config.create_sync_engine()
    elif dialect == "sqlite":
        # In memory sqlite database tables only persist within the same
        # database connection used to create them. Using a 'StaticPool'
        # ensures the same connection is always reused and
        # 'check_same_thread=False' allows the same connection to be reused
        # across threads (if necessary)
        config = DatabaseConnectionConfig(
            drivername="sqlite",
            database=":memory:",
            poolclass_name="StaticPool",
            pool_size=None,
            max_overflow=None,
        )

        engine = config.create_sync_engine()
    else:
        raise ValueError(f"Unsupported dialect: {dialect}")

    ScrachyBase.metadata.create_all(bind=engine)

    try:
        yield engine
    finally:
        # Catching exceptions here will hide any errors and potentially even
        # assertion failures
        ScrachyBase.metadata.drop_all(bind=engine)
        engine.dispose()



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

# Standard Library
import logging

from contextlib import AbstractContextManager
from contextlib import nullcontext as does_not_raise
from typing import Any, Generator, Optional
from unittest import mock

# 3rd Party Library
import pytest

from scrapy.crawler import Crawler
from scrapy.http.request import Request
from scrapy.http.response.html import HtmlResponse
from scrapy.settings import Settings, iter_default_settings
from selenium.webdriver.common.by import By
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.support import expected_conditions
from selenium.webdriver.support.wait import WebDriverWait

# 1st Party Library
from scrachy.http_ import SeleniumRequest
from scrachy.middleware.selenium import SeleniumMiddleware
from scrachy.settings.defaults import selenium as selenium_defaults
from tests.integration.middleware.conftest import MockSpider, SettingsFactory

logging.getLogger("selenium").setLevel(logging.INFO)
logging.getLogger("urllib3").setLevel(logging.INFO)
log = logging.getLogger("test_selenium")


def paginate(driver: WebDriver, request_: Request) -> dict[str, Any]:
    responses = dict()
    links = driver.find_elements(By.CLASS_NAME, "year-link")

    for link in links:
        link_text = link.text
        link.click()

        table = driver.find_element(By.CSS_SELECTOR, "table.table")
        WebDriverWait(driver, 5).until(expected_conditions.visibility_of(table))

        responses[link_text] = HtmlResponse(
            url=request_.url,
            body=driver.page_source,
            encoding="utf-8",
            request=request_,
        )

    return responses


@pytest.fixture(scope="function")
def middleware_psutils() -> Generator[SeleniumMiddleware, None, None]:
    settings = Settings(dict(iter_default_settings()))
    settings.setmodule(selenium_defaults)
    settings.set("SCRACHY_SELENIUM_WEB_DRIVER", "Chrome")
    settings.set("SCRACHY_SELENIUM_WEB_DRIVER_OPTIONS", ["--headless=new"])

    crawler = Crawler(spidercls=MockSpider, settings=settings)
    middleware = SeleniumMiddleware.from_crawler(crawler)

    yield middleware

    middleware.spider_closed(crawler.spider)


@pytest.fixture(scope="function")
def middleware_native(monkeypatch) -> Generator[SeleniumMiddleware, None, None]:
    # Standard Library
    import sys

    monkeypatch.delitem(sys.modules, "psutil", raising=False)

    settings = Settings(dict(iter_default_settings()))
    settings.setmodule(selenium_defaults)
    settings.set("SCRACHY_SELENIUM_WEB_DRIVER", "Chrome")
    settings.set("SCRACHY_SELENIUM_WEB_DRIVER_OPTIONS", ["--headless=new"])

    crawler = Crawler(spidercls=MockSpider, settings=settings)
    middleware = SeleniumMiddleware.from_crawler(crawler)

    yield middleware

    with mock.patch("scrachy.middleware.selenium.psutil", None):
        middleware.spider_closed(crawler.spider)


@pytest.mark.parametrize(
    "driver, options, extensions, raises_expectation",
    [
        (None, None, None, does_not_raise()),
        ("Chrome", ["--headless=new"], True, does_not_raise()),
        ("Chrome", ["--headless=new"], True, does_not_raise()),
        ("Firefox", ["-headless"], True, does_not_raise()),
    ],
    indirect=["extensions"],
)
def test_initialize_driver(
    settings_factory: SettingsFactory,
    driver: Optional[str],
    options: Optional[list[str]],
    extensions: Optional[list[str]],
    raises_expectation: AbstractContextManager,
):
    with raises_expectation:
        # Ignore the error, because that's what we are testing here
        settings = settings_factory(driver, options, extensions)  # type: ignore

        crawler = Crawler(spidercls=MockSpider, settings=settings)

        try:
            middleware = SeleniumMiddleware.from_crawler(crawler)
            assert middleware.driver is not None
        except:  # noqa
            middleware = None
        finally:
            if middleware is not None:
                middleware.spider_closed(crawler.spider)


def test_return_none_on_scrapy_request(middleware_psutils: SeleniumMiddleware):
    scrapy_request = Request(url="http://not-an-url")

    assert middleware_psutils.process_request(scrapy_request) is None


def test_return_response(middleware_psutils: SeleniumMiddleware):
    request = SeleniumRequest(url="https://www.scrapethissite.com/pages/forms/")
    response = middleware_psutils.process_request(request)

    assert response is not None and isinstance(response, HtmlResponse)

    title = response.css("title::text").get()

    assert title is not None
    assert "Hockey Teams" in title


def test_return_response_native(middleware_native: SeleniumMiddleware):
    # Check to make sure the clean up code using native linux commands works.
    # We only really need to test this once, with any function that uses the
    # middleware.
    request = SeleniumRequest(url="https://www.scrapethissite.com/pages/forms/")
    response = middleware_native.process_request(request)

    assert response is not None and isinstance(response, HtmlResponse)

    title = response.css("title::text").get()

    assert title is not None
    assert "Hockey Teams" in title


def test_screenshot(middleware_psutils: SeleniumMiddleware):
    request = SeleniumRequest(
        url="https://www.scrapethissite.com/pages/forms/", screenshot=True
    )
    response = middleware_psutils.process_request(request)

    assert response is not None and isinstance(response, HtmlResponse)

    assert response.meta["screenshot"] is not None


def test_script(middleware_psutils: SeleniumMiddleware):
    request = SeleniumRequest(
        url="https://www.scrapethissite.com/pages/ajax-javascript/",
        script_executor=paginate,
    )

    response = middleware_psutils.process_request(request)
    expected_titles = {
        "2015": "Spotlight",
        "2014": "Birdman",
        "2013": "12 Years a Slave",
        "2012": "Argo",
        "2011": "The Artist",
        "2010": "The King's Speech",
    }

    assert response is not None

    for key, script_response in response.meta["script_result"].items():
        title = (
            script_response.css("tbody#table-body > tr.film > td.film-title::text")
            .get()
            .strip()
        )

        assert expected_titles[key] == title
        assert expected_titles[key] == title



#  Copyright 2020 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""The SqlAlchemy models."""

# Future Library
from __future__ import annotations

# Standard Library
from typing import Optional, dataclass_transform

# 3rd Party Library
import sqlalchemy

from rwskit.sqlalchemy.base import base_mapped_column as mapped_column
from sqlalchemy import ColumnElement, ForeignKey
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Mapped, relationship
from sqlalchemy_utils import ArrowType

# 1st Party Library
from scrachy.db.base import ScrachyBase, binary, conditional_json, smallint, timestamp
from scrachy.utils.sqltypes import ConditionalJson


# Primary tables
@dataclass_transform(kw_only_default=True)
class Response(ScrachyBase, kw_only=True):
    """
    Stores the minimum amount of information to recreate the html of a
    page given its request fingerprint. The cache only stores the most
    recent version of the page, but the :class:`ScrapyHistory` table
    can be used to view

    Additional data is stored in other tables, but can be accessed via
    relationships.
    """

    fingerprint: Mapped[binary] = mapped_column(primary_key=True, natural_key=True)
    """A hash of the page's uniquely identifying request information."""

    scrape_timestamp: Mapped[timestamp] = mapped_column(ArrowType, default=None)
    """The most recent timestamp when the page was scraped."""

    url: Mapped[str] = mapped_column(index=True, compare=False)
    """
    This is redundant with the request, but can prevent a join when
    we might want to search or filter responses by it.
    """

    method: Mapped[bool] = mapped_column(default=None, compare=False)
    """The request method. ``True`` for GET. ``False`` for POST"""

    request_body: Mapped[Optional[binary]] = mapped_column(default=None, compare=False)
    """The body of the request (primarily used for POST requests)"""

    body: Mapped[binary] = mapped_column(default=None, compare=False)
    """The response body"""

    meta: Mapped[Optional[conditional_json]] = mapped_column(
        ConditionalJson, default=None, compare=False
    )
    """Store the metadata associated with the response as JSON."""

    status: Mapped[Optional[smallint]] = mapped_column(default=None, compare=False)
    """The response status code for the page"""

    headers: Mapped[Optional[str]] = mapped_column(default=None, compare=False)
    """The response headers"""

    extracted_text: Mapped[Optional[str]] = mapped_column(default=None, compare=False)
    """
    The text of the response stripped of HTML and possibly stripped of
    boilerplate content.
    """

    body_length: Mapped[int] = mapped_column(compare=False)
    """The number of bytes of the full downloaded response (including the HTML)."""

    extracted_text_length: Mapped[Optional[int]] = mapped_column(
        default=None, compare=False
    )
    """The number of bytes of the response excluding the HTML."""

    scrape_history: Mapped[list[ScrapeHistory]] = relationship(
        cascade="all, delete-orphan", default_factory=list, repr=False, compare=False
    )
    """The scrape history"""

    # I believe hybrid properties allow changing how the property behaves
    # depending on whether you are using it as an instance or as a class
    # template. For example, when operating on an instance referencing the
    # `body` property will use the regular python code in the method body.
    # But if you using it as a template for SQL operations you can define
    # a @xxx.inplace.expression method that will use the same property name
    # but emit SQL code instead.

    # So here, for example, it uses less space to store the request method
    # as a boolean (because it can only ever have two values), but it is
    # more convenient to work with these values as their original string
    # representations.
    @hybrid_property
    def request_method(self) -> Optional[str]:  # type: ignore
        """
        Get the request method as a string instead of the stored boolean type.
        """
        if self.method is True:
            return "GET"
        elif self.method is False:
            return "POST"

        return None

    @request_method.inplace.setter  # type: ignore
    def request_method(self, method: str) -> None:  # type: ignore
        """
        Client side setter.

        :param method:
        :return:
        """
        self.method = method.upper() == "GET"  # noqa

    # noinspection PyMethodParameters,PyNestedDecorators
    @request_method.inplace.expression
    @classmethod
    def request_method(cls) -> ColumnElement[Optional[str] | None]:
        """
        Server side setter.

        :return:
        """
        # return type_coerce(
        #     sqlalchemy.case(
        #         (cls.method == True, "GET"), (cls.method == False, "POST"), else_="GET"
        #     ),
        #     String,
        # )
        return sqlalchemy.case(
            (cls.method == True, "GET"),  # noqa: E712
            (cls.method == False, "POST"),  # noqa: E712
        )


@dataclass_transform(kw_only_default=True)
class ScrapeHistory(ScrachyBase, kw_only=True):
    """
    Store the history of when a page was scraped.
    """

    fingerprint: Mapped[binary] = mapped_column(
        ForeignKey("response.fingerprint"), primary_key=True, natural_key=True
    )

    # When the page was scraped
    scrape_timestamp: Mapped[timestamp] = mapped_column(
        ArrowType, primary_key=True, natural_key=True
    )

    # The response body
    body: Mapped[binary] = mapped_column(compare=False)



#  Copyright 2023 Reid Swanson.
#
#  This file is part of scrachy.
#
#  scrachy is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  scrachy is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#   along with scrachy.  If not, see <https://www.gnu.org/licenses/>.

"""
The default settings for configuring
:class:`~scrachy.utils.request.DynamicHashRequestFingerprinter`.
"""

# Standard Library
from typing import Type

# 1st Party Library
from scrachy.utils.hash import Hasher

SCRACHY_REQUEST_FINGERPRINTER_HASHER_CLASS: str | Type[Hasher] = "hashlib.sha1"
"""
The hash algorithm to use for fingerprinting the :class:`scrapy.http.Request`.
It can either be specified as an import path to a class implementing the
:class:`~scrachy.utils.hash.Hasher` protocol or a ``Hasher`` class directly.
"""
